#!/usr/bin/perl
###
### http://bach.istc.kobe-u.ac.jp/metcha/
### Metcha Search Engine 2.4
### Copyright: Naoyuki Tamura (tamura@kobe-u.ac.jp)
### 1999/06/25 version 1.0 released
### 1999/07/03 version 1.1 released
### 2000/03/17 version 1.2 released
### 2000/10/25 version 1.3 released
### 2001/01/31 version 1.4 released
### 2001/05/07 version 1.5 released
### 2001/09/18 version 1.6 released
### 2001/12/20 version 1.7 released
### 2003/04/21 version 2.0 released
### 2004/09/15 version 2.1 released
### 2006/01/06 version 2.2 released
### 2008/01/07 version 2.3 released
### 2008/12/22 version 2.4 released
### This program is distributed as a free software for non-profit use.
###
# use strict qw(vars subs);
use POSIX;
use Socket;
use Getopt::Std;
# use Jcode;
require 'jcode.pl';

use vars qw(%const %param %cgi_param);

# %eng : extract sock status time content length count sum max min
use vars qw(%eng);
use vars qw($good);
use vars qw(%good %site_max %score %rank %total_rank %word);
use vars qw(%title %comment %total);
use vars qw($opt_h $opt_e $opt_s $opt_m $opt_d $opt_S
	    $opt_C $opt_t $opt_b $opt_D);

&main;

################################################################
##
## Initialize Constants and Parameters
##
################################################################
##
## Initialize Constants
##
sub init_consts {
    $| = 1;		# No buffering
    if (%const) {
	return;
    }
    # $const{'log'} = '/home/httpd/cgi-bin/metcha.log';
    $const{'log'} = '/home/httpd/cgi-bin/log/metcha%.log';
    # $const{'logmod'} = 0666;
    $const{'toppage'} = 'http://bach.istc.kobe-u.ac.jp/metcha/';
    $const{'bmtoppage'} = 'http://bach.istc.kobe-u.ac.jp/metcha/search.html';
    $const{'metcha'} = 'http://bach.istc.kobe-u.ac.jp/cgi-bin/metcha.cgi';
    $const{'helppage'} = 'http://bach.istc.kobe-u.ac.jp/metcha/help.html';
    $const{'bmpage'} = 'http://bach.istc.kobe-u.ac.jp/metcha/bookmarks.html';
    $const{'bmscript'} = 'http://bach.istc.kobe-u.ac.jp/metcha/bookmarks.js';
    $const{'maintainer'} = 
	'<a href="http://bach.istc.kobe-u.ac.jp/tamura-jp.html" target="_top">Naoyuki Tamura</a>';
    $const{'from'} = 'tamura@kobe-u.ac.jp';
    # $const{'re_allow_ip'} = '^133\.30\.';
    # $const{'proxy_server'} = 'proxy.xxx.yy.jp';
    # $const{'proxy_port'} = 8080;
    # $const{'no_proxy'} = 'xxx\.yy\.jp\$';
    $const{'max_age'} = 600;
    $const{'agent_name'} = "Metcha/2.4 ($const{'toppage'})";
    $const{'programmer'} =
	'<a href="http://bach.istc.kobe-u.ac.jp/tamura-jp.html" target="_top">'
	    . 'Naoyuki Tamura</a>';
    # All engines
    $const{'engines0'} =
	[ sort(
#	       'altavista', 'excite', 'fast', 'fresheye', 'infoseek', 'verno',
	       'excite', 'fresheye', 'goo', 'yahoo',
	       ) ];
    # Default engines
    $const{'engines1'} =
	[ sort(
#	       'altavista', 'excite', 'fast', 'fresheye', 'infoseek', 'verno',
	       'excite', 'fresheye', 'goo', 'yahoo',
	       ) ];
}

##
## Initialize Parameters
##
sub init_params {
    %param = ();
    my $t0 = time;
    @_ = localtime($t0);
    my $ym = sprintf "%04d%02d", $_[5] + 1900, $_[4] + 1;
    $param{'start_time'} = $t0;
    $param{'id'} = "$t0:$$";
    $param{'current_date'} = scalar(localtime($t0));

    my $logfile = $const{'log'};
    $logfile =~ s/%/$ym/;
    $param{'logfile'} = $logfile;

    my (@engines, $query, $speed, $max_show, $domain, $list_by_site,
	$with_comment, $target_win, $use_bm, $word, $debug);
#    my ($opt_h, $opt_e, $opt_s, $opt_m, $opt_S,
#	$opt_C,	$opt_t, $opt_b, $opt_D);

    if (! $ENV{'REQUEST_METHOD'}) {
	# Command line
	$param{'cgi'} = 0;
	&getopts("he:s:m:d:Sc:t:bw:D:");
	if ($opt_h || @ARGV == 0) {
	    print "Usage: $0 [options] word ...\n";
	    print "  -h         : Help\n";
	    print "  -e engines : Engines to use\n";
	    printf "\t\t(%s)\n", join(',', @{$const{'engines0'}});
	    print "  -s {f,m,s} : Speed (fast, medium, slow)\n";
	    print "  -m n       : Max number of items (0 means all)\n";
	    print "  -d domain  : Restrict domain\n";
	    print "  -S         : List by site\n";
	    print "  -c {-1,0,1}: Comments (short, medium, long)\n";
	    print "  -t target  : Target window name\n";
	    print "  -b         : Enable Bookmarks\n";
	    print "  -w n       : Enable word suggestions\n";
	    print "  -D {s,l}   : Debug (save, load)\n";
	    return undef;
	}
	$query = join(' ', @ARGV);
	if ($opt_e) {
	    my $e;
	    foreach $e (split(/,/, $opt_e)) {
		next unless grep($e eq $_, @{$const{'engines0'}});
		push(@engines, $e);
	    }
	}
	$speed = $opt_s;
	$max_show = $opt_m;
	$domain = $opt_d;
	$list_by_site = $opt_S;
	$with_comment = $opt_c;
	$target_win = $opt_t;
	$use_bm = $opt_b;
	$word = $opt_w;
	$debug = $opt_D;
    } else {
	# CGI
	$param{'cgi'} = 1;
	&cgi_params;
	$query = $cgi_param{'q'};
	@engines = grep($cgi_param{$_}, @{$const{'engines0'}});
	$speed = $cgi_param{'speed'};
	$max_show = $cgi_param{'max'};
	$domain = $cgi_param{'domain'};
	$list_by_site = $cgi_param{'site'};
	$with_comment = $cgi_param{'comment'};
	$target_win = $cgi_param{'tw'};
	$use_bm = $cgi_param{'bm'};
	$word = $cgi_param{'word'};
	$param{'remote_host'} = $ENV{'REMOTE_HOST'} || '?';
	$param{'remote_addr'} = $ENV{'REMOTE_ADDR'} || '?';
    }

    $query = '' unless defined $query;
    &jcode::convert(\$query, 'euc', undef, 'z');
    &jcode::tr(\$query, "\xa1\xa1", " ");
    $query = &remove_strange_chars($query);
    $query =~ s/^\s+//;
    $query =~ s/\s+$//;
    $query =~ s/\s+/ /;
    $param{'q'} = $query;
    $param{'q_euc'} = &url_encode($query);
    $param{'q_sjis'} = &url_encode(&jcode::sjis($query, 'euc'));

    if (@engines == 0) {
	@engines = @{$const{'engines1'}};
    }
    @engines = sort @engines;
    $param{'engines'} = \@engines;

    $speed = 'f' unless defined $speed;
    if ($speed eq 'm') {
	$param{'max_hits'} = 30;
	$param{'send_timeout'} = 5;
	$param{'recv_timeout'} = 20;
    } elsif ($speed eq 's') {
	$param{'max_hits'} = 50;
	$param{'send_timeout'} = 10;
	$param{'recv_timeout'} = 30;
    } else {
	$speed = 'f';
	$param{'max_hits'} = 10;
	$param{'send_timeout'} = 3;
	$param{'recv_timeout'} = 10;
    }
    $param{'speed'} = $speed;
    $param{'extract_timeout'} = 3;

    $max_show = (defined $max_show && $max_show >= 0) ? int($max_show) : 50;
    $param{'max'} = $max_show;

    $domain = '' unless defined $domain;
    if ($domain) {
	&jcode::convert(\$domain, 'euc', undef, 'z');
	&jcode::tr(\$domain, "\xa1\xa1", " ");
	&jcode::tr(\$domain, "ݡ-ڣ--", "-_./A-Za-z0-9");
	$domain = &remove_strange_chars($domain);
	$domain =~ s/\s+//g;
	$domain =~ s|^http://||;
	$domain =~ s|^([^/]*)/.*$|$1|;
    }
    $param{'domain'} = $domain || '';

    $list_by_site = 0 unless defined $list_by_site;
    $param{'site'} = int($list_by_site);

    $with_comment = 0 unless defined $with_comment;
    $param{'comment'} = int($with_comment);

    $target_win = '' unless defined $target_win;
    $target_win =~ tr/a-zA-Z0-9_//cd;
    $target_win = '_top' unless $target_win;
    $param{'tw'} = $target_win;

    if ($use_bm && $const{'bmpage'} && $const{'bmscript'}) {
	$param{'bm'} = 1;
    } else {
	$param{'bm'} = 0;
    }

    $param{'word'} = int($word);
    if ($param{'word'} < 0) {
	$param{'word'} = 5;
    }
    
    $param{'debug'} = $debug || '';

    my $menu;
    if ($param{'bm'}) {
	$menu = "<a href=\"$const{'bmtoppage'}\">ȥåץڡ</a>";
	$menu .= " | <a href=\"$const{'helppage'}\" target=\"_top\">إ</a>";
	$menu .= " | <a href=\"$const{'bmpage'}\" target=\"BOOKMARKS\">";
	$menu .= "ɽ</a>";
	$menu .= " | <a href=\"javascript:top.location.href=location.href\">";
	$menu .= "ե졼</a>";
    } else {
	$menu = "<a href=\"$const{'toppage'}\">ȥåץڡ</a>";
	$menu .= " | <a href=\"$const{'helppage'}\" target=\"_top\">إ</a>";
    }
    $param{'menu'} = $menu;

    return 1;
}

##
## Get CGI parameters
##
sub cgi_params {
    my ($args, $tag, $value);
    if ($ENV{'REQUEST_METHOD'} eq 'POST'){
	sysread(STDIN, $args, $ENV{'CONTENT_LENGTH'});
    } else {
	$args = $ENV{'QUERY_STRING'};
    }
    foreach (split('&', $args)) {
	$tag = ''; $value = '';
	($tag, $value) = split(/=/, $_, 2);
	$value =~ s/\+/ /g;
	$value =~ s/%(..)/pack('c', hex($1))/ge;
	$cgi_param{$tag} = $value;
    }
}

################################################################
##
## Main
##
################################################################
sub main {
    &init_consts;
    if (! &init_params) {
	return;
    }
    my $need_resp = $ENV{'MOD_PERL'} && ! $ENV{'PERL_SEND_HEADER'};
    my $query = $param{'q'};
#    if ($query =~ /^\s*$/) {
#	if ($need_resp) {
#	    print "HTTP/1.1 302 Found\r\n";
#	}
#	print "Location: $const{'toppage'}\r\n\r\n";
#	return;
#    }
    if ($need_resp) {
	print "HTTP/1.1 200 OK\r\n";
    }
    &print_header;
    if ($param{'cgi'}) {
	&write_log_1;
	my $re_allow_ip = $param{'re_allow_ip'};
	if ($re_allow_ip && $ENV{'REMOTE_ADDR'} !~ /$re_allow_ip/) {
	    print "You are not authorized to use this service.\n";
	    print "</body></html>\n";
	    return;
	}
    }

    if ($query =~ /^\s*$/) {
	&print_footer;
	return;
    }

    &body;

    &print_footer;			# Print footer
    if ($param{'cgi'}) {
	&write_log_2;
    }
}

sub body {
    my $eng;

    my @engines = @{$param{'engines'}};
	
    ## Generate requests
    print "<hr>\n";
    print "Searching\n";
    foreach $eng (@engines) {
	&init_engine($eng, $param{'max_hits'});
	my $req = &html_encode($eng{$eng}{'request'});
	my $href = "http://$eng{$eng}{'server'}$req";
	my $target = defined $param{'tw'} ? " target=\"$param{'tw'}\"" : '';
	print "[<a href=\"$href\"$target>";
	print "$eng</a>]\n";
    }
    print "<br>\n";

    ## Send requests
    my $n = @engines;
    print "<font size=-1 color=\"#808080\">";
    print "Connecting to $n engines (seconds):\n";
    foreach $eng (@engines) {
	my $t0 = time;
	if ($param{'debug'} eq 'l') {
	    $eng{$eng}{'status'} = '';
	} else {
	    &request($eng);
	}
	my $t = time - $t0;
	if ($eng{$eng}{'status'} ne 'err') {
	    print "$eng=$t\n";
	} else {
	    print "$eng=x\n";
	}
    }
    print "</font><br>\n";

    ## Receive from all engines
    print "<font size=-1 color=\"#808080\">";
    $_ = grep(($eng{$_}{'status'} ne 'err'), @engines);
    print "Receiving from $_ engines (seconds):\n";
    if ($param{'debug'} eq 'l') {
	foreach $eng (@engines) {
	    my $file = "$eng.log";
	    &load_content($eng, $file);
	    $eng{$eng}{'status'} = 'ok';
	    print "$eng=0\n";
	}
    } else {
	&receive_all(time);
    }
    print "</font><br>\n";
    if ($param{'debug'} eq 's') {
	foreach $eng (@engines) {
	    my $file = "$eng.log";
	    unlink $file;
	    next if $eng{$eng}{'status'} eq 'err';
	    &save_content($eng, $file);
	}
    }

    ## Extract URLs
    %good = ();
    %site_max = ();
    %score = ();
    %rank = ();
    %total_rank = ();
    %title = ();
    %comment = ();
    %total = ();
    print "<font size=-1 color=\"#808080\">";
    $_ = grep(($eng{$_}{'status'} ne 'err'), @engines);
    print "Extracting from $_ engines (URLs):\n";
    foreach $eng (@engines) {
	next if $eng{$eng}{'status'} eq 'err';
	&extract($eng);
	print "$eng=$eng{$eng}{'count'}\n";
	undef $eng{$eng}{'content'};
    }
    print "</font><br>\n";

    ## Calculate total score
    &calc_total;
    $eng = 'TOTAL';
    $eng{$eng}{'status'} = 'ok';
    $eng{$eng}{'time'} = time - $param{'start_time'};
    print ": $eng{$eng}{'count'} ڡ ($eng{$eng}{'time'} )<br>\n";

    ## Suggest words
    if ($param{'word'}) {
	my $query = $param{'q'};
	my %word_count = ();
	foreach my $w (keys %word) {
	    next if length($w) <= 2;
	    next if index($query, $w) >= 0;
	    my $c = scalar(keys(%{$word{$w}}));
	    if ($c > 1) {
		$word_count{$w} = $c;
	    }
	}
	my @w = sort { $word_count{$b} <=> $word_count{$a} } keys %word_count;
	@w = splice(@w, 0, $param{'word'});
	foreach my $w (@w) {
	    my $c = $word_count{$w};
	    print "$w ($c) ";
	}
	print "\n";
    }

    print "<hr>\n";

    ## Order by the total score
    my @hrefs = sort by_total keys %total;
    my $max_show = $param{'max'};
    if ($max_show > 0 && @hrefs > $max_show) {
	splice(@hrefs, $max_show);
    }
    my $c = 0;
    my $href;
    foreach $href (@hrefs) {
	$total_rank{$href} = ++$c;
    }

    ## List Pages
    print "<form name=\"results\" action=\"javascript:void(1)\">\n";

    ## List Good Pages
    $good = 0;
    if (%good) {
	print "<!-- begin good pages -->\n";
	&list_good_pages;
	print "<!-- end good pages -->\n";
	print "<hr>\n";
    }

    ## List Pages
    print "<!-- begin results -->\n";
    if (! $param{'site'}) {
	# List by page
	&list_by_page(@hrefs);
    } else {
	# List by site
	&list_by_site(@hrefs);
    }
    print "<!-- end results -->\n";
    print "</form>\n";

    ## Print links
    &print_links;
}

################################################################
##
## Print HTML Page
##
################################################################
##
## Print Header
##
sub print_header {
    my $title0 = "Metcha Search for: ";
    my $q = &html_encode($param{'q'});
    my (%sel_speed, %sel_max, %sel_site, %sel_comment, $domain);
    foreach ('f', 'm', 's') { $sel_speed{$_} = ''; }
    $sel_speed{$param{'speed'}} = ' selected';
    foreach (10, 20, 50, 100, 0) { $sel_max{$_} = ''; }
    $sel_max{$param{'max'}} = ' selected';
    foreach (0, 1) { $sel_site{$_} = ''; }
    $sel_site{$param{'site'}} = ' selected';
    foreach (-1, 0, 1) { $sel_comment{$_} = ''; }
    $sel_comment{$param{'comment'}} = ' selected';
    $domain = &html_encode($param{'domain'});
    my ($bmmeta, $bmsrc, $bmbody, $bmwarning) = ('', '', '', '');
    if ($param{'bm'}) {
	$bmmeta = '<meta http-equiv="Content-Script-Type" '
	    . 'content="text/javascript">' . "\n";
	$bmsrc = &javascript($const{'bmscript'}, "") . "\n";
#	$bmbody = " onload=\"refreshCheck(document.results)\""
#	    . " onfocus=\"refreshCheck(document.results)\"";
#	$bmbody = " onload=\"refreshCheckMarks\""
#	    . " onfocus=\"refreshCheckMarks\"";
	$bmwarning = "<noscript><p><font color=\"#ff0000\">"
	    . "JavaScriptCookie򥪥ˤʤȤ굡ǽѤǤޤ"
	    . "</font></p></noscript>\n";
    }
    if ($param{'cgi'}) {
	print "Content-type: text/html\r\n";
	if ($const{'max_age'}) {
	    print "Cache-Control: max-age=$const{'max_age'}, ",
	          "must-revalidate\r\n";
	}
	print "\r\n";
    }
    $_ = <<EOM
<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN">
<html lang="ja">
<head>
<!-- $const{'agent_name'} -->
<meta http-equiv="Content-Type" content="text/html; charset=EUC-JP">
$bmmeta<title>$title0$q</title>
$bmsrc</head>
<body bgcolor="#ffffff"$bmbody>
<table border=0 cellpadding=0 cellspacing=0 bgcolor="#000060" width="100%" summary="header">
<tr><td>&nbsp;&nbsp;<font color="#80ffff"><font face="sans-serif">$title0</font>$q</font></td></tr>
</table>
<br>
[ $param{'menu'} ]
<br>
<form name="metcha" method="get" action="$const{'metcha'}">
<input size=50 name="q" value="$q"><input type="submit" value="">
<br>
<select name="speed">
<option value="f"$sel_speed{'f'}>/®
<option value="m"$sel_speed{'m'}>/®
<option value="s"$sel_speed{'s'}>/®
</select>
<select name="max">
<option value="10"$sel_max{10}>10ɽ
<option value="20"$sel_max{20}>20ɽ
<option value="50"$sel_max{50}>50ɽ
<option value="100"$sel_max{100}>100ɽ
<option value="0"$sel_max{0}>٤ɽ
</select>
<select name="site">
<option value="0"$sel_site{0}>ڡ
<option value="1"$sel_site{1}>Ȥ
</select>
<select name="comment">
<option value="-1"$sel_comment{0}>ʸʤ
<option value="0"$sel_comment{0}>ʸ̾
<option value="1"$sel_comment{1}>ʸ
</select>
EOM
    ;
    if ($domain) {
	$_ .= <<EOM
<br>
ɥᥤ <input size=35 name="domain" value="$domain">
EOM
    }
    print;
    foreach (@{$param{'engines'}}) {
	print "<input type=hidden name=$_ value=1>\n";
    }
    $_ = <<EOM
<input type=hidden name=tw value=\"$param{'tw'}\">
<input type=hidden name=bm value=$param{'bm'}>
$bmwarning
</form>
EOM
    ;
    print;
}

##
## JavaScript block
##
sub javascript {
    my ($src, $prog) = @_;
    if ($src) {
	$src = " src=\"$src\"";
    }
#    "<script language=\"javascript\"$src>$prog</script>";
    "<script type=\"text/javascript\"$src>$prog</script>";
}

##
## Print Links
##
sub print_links {
    my $q_euc = $param{'q_euc'};
    my $q_sjis = $param{'q_sjis'};
    my $domain = &url_encode($param{'domain'});
    my (%links, $q, $eng, $link);
    $q = $q_sjis;
    if ($domain) {
	$q .= '+%2Bsite:' . $domain;
    }
    $links{'google'} = 'http://www.google.com/search?'
	. 'q=' . $q . '&amp;hl=ja&amp;ie=Shift_JIS';
    $q = $q_euc;
    if ($domain) {
	$q .= '+%2Bsite:' . $domain;
    }
    $links{'yahoo'} = 'http://search.yahoo.co.jp/search/dir?'
	. 'p=' . $q;
#    $links{'dmoz'} = 'http://search.dmoz.org/cgi-bin/search?'
#	. 'search=' . $q_sjis . '&all=yes&cs=Shift_JIS&cat=World%2FJapanese';
#    $links{'lycos'} = 'http://search.lycos.co.jp/main.html?'
#	. 'query=' . $q_sjis . '&amp;encoding=shift-jis';
#    $links{'vivisimo'} = 'http://vivisimo.com/search?'
#	. 'query=' . $q_sjis;
    $links{'mars'} = 'http://www.marsflag.com/?mode=web&detrimental=all'
	. '&imgsize=auto&mode=web&key=' . $q_euc;

    print "<hr>\n";
    print "<font size=-1>¾θ󥸥Ǹ: ";
    foreach $eng (sort keys %links) {
	print "[<a href=\"$links{$eng}\" target=\"_top\">$eng</a>] ";
    }
    print "</font>\n";
}

##
## Print Footer
##
sub print_footer {
    if ($param{'bm'}) {
#	print &javascript("", "refreshCheck(document.results);") . "\n";
	print &javascript("", 
			  "window.onload = refreshCheckMarks;" .
			  "window.onfocus = refreshCheckMarks;") . "\n";
    }
    print "<hr>\n";
    print "[ $param{'menu'} ]\n";
    print "<hr>\n";
    print "<address>\nMaintained by $const{'maintainer'} / ";
    print "Programmed by $const{'programmer'}\n</address>\n";
    print "</body>\n</html>\n";
}

##
## List Good Pages
##
sub list_good_pages {
    my ($href);
    my $target = defined $param{'tw'} ? " target=\"$param{'tw'}\"" : '';
    print "<ul>\n";
    foreach $href (sort keys %good) {
	my @e = sort keys %{$good{$href}};
	next if @e == 0;
	my $title = $good{$href}{$e[0]};
	$title = $href unless $title;
	my $e = join(', ', @e);
	print "<li> ڡ: ";
	&checkbox("g$good", $href, $title);
	print "<a href=\"$href\"$target>$title</a> ";
	print "<font size=-1 color=\"#808080\">($e)</font>\n";
	++$good;
    }
    print "</ul>\n";
}

##
## List by page
##
sub list_by_page {
    my @hrefs = @_;
    my $href;
    print "<dl>\n";
    foreach $href (@hrefs) {
	&list_href($href);
    }
    print "</dl>\n";
}

##
## List by site
##
sub list_by_site {
    my @hrefs = @_;
    my ($site, @site_hrefs, $href);
    print "<dl>\n";
    foreach $site (sort by_site_max keys %site_max) {
	@site_hrefs = ();
	foreach $href (@hrefs) {
	    next if $href !~ m|^http://([\-\.\w]+)|i;
	    next if $1 ne $site;
	    push(@site_hrefs, $href);
	}
	next if @site_hrefs == 0;
	print "<dt> <i>$site</i>\n";
	print "<dd>\n";
	print "<dl>\n";
	foreach $href (@site_hrefs) {
	    &list_href($href);
	}
	print "</dl>\n";
    }
    print "</dl>\n";
}

##
## List URL
##
sub list_href {
    my ($href) = @_;
    my (@e, $eng);
    foreach $eng (sort keys %{$score{$href}}) {
	push(@e, "$eng#$rank{$href}{$eng}=$score{$href}{$eng}");
    }
    my $i = $total_rank{$href};
    my $title = $title{$href};
    if (! $title || length($title) <= 1) {
	$title = $href;
    }
    my $comment = ($comment{$href} || $title);
    my $total = int($total{$href});
    my $target = defined $param{'tw'} ? " target=\"$param{'tw'}\"" : '';
    # $ml = "$metchalink?id=$id";
    # $ml .= "&url=" . &url_encode($href);
    print "<dt> $i ";
    print "<font size=-1 color=\"#ff0000\">[$total]</font> ";
    &checkbox("p$i", $href, $title);
    print "<a href=\"$href\"$target>$title</a>\n";
    if ($param{'comment'} == 1) {
	print "<dd> <font size=-1>$comment</font>\n";
    }
    if ($param{'comment'} != -1) {
	my $e = join(', ', @e);
	print "<dd> <font size=-1 color=\"#808080\">($e) <i>$href</i></font>\n";
	# print "[<a href=\"$ml\">is referred from</a>]\n";
    }
}

##
## Checkbox
##
sub checkbox {
    my ($name, $href, $title) = @_;
    if ($param{'bm'}) {
#	$_ = "$href " . &url_encode($title);
#	print "<input type=checkbox name=\"$name\" value=\"$href $title\"> ";
	print "<input type=checkbox name=\"$name\" ",
		"onclick=\"toBM(this)\" onkeypress=\"toBM(this)\" ",
	        "value=\"$href $title\"> ";
#		"onclick=\"toBM(this)\" value=\"$_\"> ";
    }
}

sub by_site_max {
    ($site_max{$b} <=> $site_max{$a}) || ($a cmp $b);
}

sub by_total {
    ($total{$b} <=> $total{$a}) || ($a cmp $b);
}

################################################################
##
## Send Request
##
################################################################
##
## Request to a search engine
##
sub request {
    my ($eng) = @_;
    my $sock = "S$eng";
    my $server = ($eng{$eng}{'ip'} || $eng{$eng}{'server'});
    my $port = $eng{$eng}{'port'};
    $eng{$eng}{'sock'} = '';
    $eng{$eng}{'status'} = '';
    my $req = "GET $eng{$eng}{'request'} HTTP/1.0\015\012"
	. "Host: $eng{$eng}{'server'}\015\012"
	. "User-Agent: $const{'agent_name'}\015\012"
	. "Accept-Language: ja,en\015\012"
	. "Accept-Charset: shift_jis, euc-jp\015\012"
	. "Referer: $const{'toppage'}\015\012"
	. "From: $const{'from'}\015\012\015\012";
    $SIG{'ALRM'} = \&request_timeout;
    alarm($param{'send_timeout'});
    eval {
	if (&tcp_client($sock, $server, $port)) {
	    print $sock $req;
	    $eng{$eng}{'sock'} = $sock;
	    alarm(0);
	}
    };
    alarm(0);
    $SIG{'ALRM'} = 'IGNORE';
    if (! $eng{$eng}{'sock'}) {
	$eng{$eng}{'status'} = 'err';
	close($sock);
    }
}

sub request_timeout {
    die;
}

##
## TCP client
##
sub tcp_client {
    my ($sock, $host, $port) = @_;
    my ($addr, $proto, $ent, $oldh);
    @_ = gethostbyname($host);
    $addr = $_[4];
    $proto = getprotobyname('tcp');
    if ($port !~ /^\d+$/) {
	@_ = getservbyname($port, 'tcp');
	$port = $_[2];
    }
    socket($sock, PF_INET, SOCK_STREAM, $proto);
    $ent = pack('S n a4 x8', AF_INET, $port, $addr);
    if (! connect($sock, $ent)) {
	return undef;
    }
    $oldh = select;
    select($sock); $| = 1; select($oldh);
    $sock;
}

################################################################
##
## Receive
##
################################################################
##
## Receive results from all search engines
##
sub receive_all {
    my ($time0) = @_;
    my $engines = $param{'engines'};
    my $recv_timeout = $param{'recv_timeout'};
    my ($eng, $sock);
    my ($c, $r_bitset, $nfound, $timeleft);
    my ($buf, $bytes);
    my $buflen = 8192;
    my $wait = 1;
    foreach $eng (@{$engines}) {
	$eng{$eng}{'content'} = '';
	$eng{$eng}{'time'} = '';
	$sock = $eng{$eng}{'sock'};
	next unless $sock;
	# Non blocking IO
	fcntl($sock, F_SETFL, O_NONBLOCK);
    }
    while (1) {
	$c = 0;
	$r_bitset = '';
	foreach $eng (@{$engines}) {
	    $sock = $eng{$eng}{'sock'};
	    next unless $sock;
	    next if $eng{$eng}{'status'} eq 'ok';
	    vec($r_bitset, fileno($sock), 1) = 1;
	    ++$c;
	}
	last if $c == 0;
	($nfound, $timeleft) = select($r_bitset, undef, undef, $wait);
	last if $nfound < 0;	# Error
	if ($nfound == 0) {	# Timeout
	    last if time - $time0 > $recv_timeout;
	    next;
	}
	foreach $eng (@{$engines}) {
	    $sock = $eng{$eng}{'sock'};
	    next unless $sock;
	    next if $eng{$eng}{'status'} eq 'ok';
	    next if vec($r_bitset, fileno($sock), 1) == 0;
	    $bytes = sysread($sock, $buf, $buflen);
	    if ($bytes == 0) {
		# closed
		$eng{$eng}{'status'} = 'ok';
		close($sock);
		$eng{$eng}{'time'} = time - $time0;
		print "$eng=$eng{$eng}{'time'}\n";
	    } else {
		$eng{$eng}{'content'} .= $buf;
	    }
	}
    }
    foreach $eng (@{$engines}) {
	$sock = $eng{$eng}{'sock'};
	next unless $sock;
	next if $eng{$eng}{'status'} eq 'ok';
	close($sock);
	$eng{$eng}{'status'} = 'x';	# Timeout
	$eng{$eng}{'time'} = time - $time0;
	print "$eng=x\n";
    }
}

################################################################
##
## Extract
##
################################################################
##
## Extract URLs from $eng{$eng}{'content'}
## and set the followings
##	$eng{$eng}{'length'} : length of $content{$eng}
##	$eng{$eng}{'count'}  : number of URLs
##	$eng{$eng}{'sum'}    : sum of URL scores
##	$eng{$eng}{'max'}    : max of URL scores
##	$eng{$eng}{'min'}    : min of URL scores
##	$score{$href}{$eng}  : score of URL
##	$title{$href}        : title of URL
##
sub extract {
    my ($eng) = @_;
    $_ = $eng{$eng}{'content'};
    $eng{$eng}{'length'} = length($_);
    tr/\r\n/  /;
    $eng{$eng}{'count'} = 0;
    $eng{$eng}{'sum'} = 0;
    # call the extract program of $eng
    $SIG{'ALRM'} = \&extract_timeout;
    alarm($param{'extract_timeout'});
    eval {
	&{$eng{$eng}{'extract'}}($eng, $_);
    };
    alarm(0);
    $SIG{'ALRM'} = 'IGNORE';
    $eng{$eng}{'max'} = int($eng{$eng}{'max'} || 0);
    $eng{$eng}{'min'} = int($eng{$eng}{'min'} || 0);
}

sub extract_timeout {
    die;
}

sub add_url {
    my ($eng, $href, $title, $score, $comment) = @_;
    my $domain = $param{'domain'};
    my $word = $param{'word'};
    my $code;
    $href =~ s/^\s+//;
    if ($href !~ m|^http://|) {
	return;
    }
    $href = &remove_strange_chars($href);
    $href =~ s/\s+$//;
    $href =~ s/ /%20/g;
    $href =~ s/\"/%22/g;
    $href =~ s/\'/%27/g;
    $href =~ s/</%3c/g;
    $href =~ s/>/%3e/g;
    $href =~ s/%7e|%7E/~/g;
    $href =~ s|^(http://[\w_!*\$\-\.\+\'\(\)]+)\.([:/])|$1$2|i;
    $href =~ s|^(http://[\w_!*\$\-\.\+\'\(\)]+):80\b|$1|i;
    if ($domain && $href !~ /$domain/) {
        return;
    }
    $href =~ s|/index.html?$|/|;
    if ($href =~ m|^(.*)/$|) {
	if (defined $score{$1} && ! defined $score{$href}) {
	    $score{$href} = $score{$1}; delete $score{$1};
	    $title{$href} = $title{$1}; delete $title{$1};
	    $comment{$href} = $comment{$1}; delete $comment{$1};
	    $rank{$href} = $rank{$1}; delete $rank{$1};
	    if ($good{$1}) {
		$good{$href} = $good{$1}; delete $good{$1};
	    }
	}
    } else {
	if (defined $score{"$href/"}) {
	    $href = "$href/";
	}
    }
    if (defined $score{$href}{$eng}) {
	return;
    }
    if ($score eq 'good') {
	$good{$href}{$eng} = $title;
	return;
    }
    $title =~ s|</?.*?>||gi;
    $title =~ s|&nbsp;| |gi;
    $title =~ s|\s+| |gi;
    $title =~ s/^\s+//;
    $title =~ s/\s+$//;
    if ($title) {
	$title = &html_text($title);
	$code = &jcode::getcode(\$title);
	if (! $code || $code eq 'euc') {
	    if (! $title{$href} || length($title{$href}) < length($title)) {
	        $title{$href} = $title;
            }
	}
    }
    $comment = "" unless $comment;
    $comment =~ s|</?.*?>||gi;
    $comment =~ s|&nbsp;| |gi;
    $comment =~ s|\s+| |gi;
    $comment =~ s/^\s+//;
    $comment =~ s/\s+$//;
    if ($comment) {
	$comment = &html_text($comment);
	$code = &jcode::getcode(\$comment);
	if (! $code || $code eq 'euc') {
	    if (! $comment{$href} ||
		length($comment{$href})<length($comment)) {
		$comment{$href} = $comment;
	    }
	    if ($word) {
		# print STDERR "#1 $comment\n";
		$_ = $comment;
		s/\s+/\t/g;
		$_ = " " . &insert_spaces($_);
		# print STDERR "#2 $_\n";
		my $word_regex = '(( [\xb0-\xfe][\xa1-\xfe])+|( \xa5[\xa1-\xf6]| )+)';
		while (/$word_regex/g) {
		    my $w = $1;
		    $w =~ tr/ //d;
		    $word{$w}{$href}++;
		}
	    }
	}
    }
    ++$eng{$eng}{'count'};
    $rank{$href}{$eng} = $eng{$eng}{'count'};
    $score{$href}{$eng} = $score;
    $eng{$eng}{'sum'} += $score;
    if (! $eng{$eng}{'max'} || $score > $eng{$eng}{'max'}) {
	$eng{$eng}{'max'} = $score;
    }
    if (! $eng{$eng}{'min'} || $score < $eng{$eng}{'min'}) {
	$eng{$eng}{'min'} = $score;
    }
}

##
## Calculate the total score
##
sub calc_total {
    my ($href, $eng, $total, $site);
    my @engines = @{$param{'engines'}};
    my $t = 'TOTAL';
    $eng{$t}{'count'} = 0;
    $eng{$t}{'sum'} = 0;
    foreach $href (keys %score) {
	next if scalar(keys %{$score{$href}}) == 0;
	$total = 0;
	foreach $eng (@engines) {
	    next unless $eng{$eng}{'max'};
	    next unless $score{$href}{$eng};
	    $total += ($score{$href}{$eng} / $eng{$eng}{'max'} + 0.3) / 1.3;
	}
	my $engines = grep(($eng{$_}{'status'} ne 'err'), @engines);
	$total = 100 * $total / $engines;
	$total{$href} = $total;
	++$eng{$t}{'count'};
	$eng{$t}{'sum'} += $total;
	if (! $eng{$t}{'max'} || $total > $eng{$t}{'max'}) {
	    $eng{$t}{'max'} = $total;
	}
	if (! $eng{$t}{'min'} || $total < $eng{$t}{'min'}) {
	    $eng{$t}{'min'} = $total;
	}
	if ($href !~ m|^http://([\w_!*\$\-\.\+\'\(\)]+)|i) {
	    next;
	}
	$site = $1;
	if (! $site_max{$site} || $total > $site_max{$site}) {
	    $site_max{$site} = $total;
	}
    }
    $eng{$t}{'max'} = int($eng{$t}{'max'} || 0);
    $eng{$t}{'min'} = int($eng{$t}{'min'} || 0);
}

################################################################
##
## Write log
##
################################################################
sub write_log_1 {
    my $logfile = $param{'logfile'};
    my $t = $param{'current_date'};
    my $line = "Q $param{'id'} [$t]";
    $line .= " $param{'remote_host'} $param{'remote_addr'}";
    my $eng;
    foreach $eng (@{$param{'engines'}}) {
	$line .= " $eng";
    }
    foreach (qw(speed max domain site comment tw bm q)) {
	$line .= " $_=$param{$_}";
    }
    open(LOG, ">>$logfile") || die "cannot append $logfile\n";
    flock(LOG, 2);		# exclusive lock
    seek(LOG, 0, 2);
    print LOG $line, "\n";
    close(LOG);
    if ($param{'logmod'}) { chmod $param{'logmod'}, $logfile };
}

sub write_log_2 {
    my $logfile = $param{'logfile'};
    my $t = time - $param{'start_time'};
    my $line = "R $param{'id'} [$t]";
    $line .= " $param{'remote_host'} $param{'remote_addr'}";
    my $eng;
    foreach $eng (@{$param{'engines'}}, 'TOTAL') {
	if ($eng{$eng}{'status'} eq 'err') {
	    $line .= " $eng=$eng{$eng}{'status'}";
	} else {
	    $line .= " $eng=$eng{$eng}{'status'}";
	    $line .= ",$eng{$eng}{'time'}";
	    $line .= ",$eng{$eng}{'count'}";
	    $line .= ",$eng{$eng}{'max'}-$eng{$eng}{'min'}";
	}
    }
    $line .= " good=$good";
    foreach (qw(speed max domain site comment tw bm q)) {
	$line .= " $_=$param{$_}";
    }
    open(LOG, ">>$logfile") || die "cannot append $logfile\n";
    flock(LOG, 2);		# exclusive lock
    seek(LOG, 0, 2);
    print LOG $line, "\n";
    close(LOG);
    if ($param{'logmod'}) { chmod $param{'logmod'}, $logfile };
}

################################################################
##
## Debug
##
################################################################
##
## Save content to a file for debug
##
sub save_content {
    my ($eng, $file) = @_;
    open (FILE, ">$file") || die "$0: $file $!";
    $_ = $eng{$eng}{'content'};
    # tr/\r/ /d;
    # &jcode::convert(\$_, 'euc', undef, 'z');
    print FILE $_;
    close(FILE);
}

##
## Load content from a file for debug
##
sub load_content {
    my ($eng, $file) = @_;
    if (! open (FILE, "<$file")) {
	return;
    }
    @_ = <FILE>;
    $eng{$eng}{'content'} = join('', @_);
    close(FILE);
}

################################################################
##
## Utilities
##
################################################################
##
## URL encode
##
sub url_encode {
    my ($line) = @_;
    $line =~ s/([^ \w])/'%'.uc(unpack('H2',$1))/ge;
    $line =~ s/\s/+/g;
    $line;
}

##
## HTML encode
##
sub html_encode {
    my ($line) = @_;
    $line =~ s/&/&amp;/g;
    $line =~ s/</&lt;/g;
    $line =~ s/>/&gt;/g;
    $line =~ s/"/&quot;/g;
    $line;
}

##
## HTML text
##
sub html_text {
    my ($line) = @_;
    $line = &remove_strange_chars($line);
    $line =~ s/^\s+//;
    $line =~ s/\s+$//;
    $line =~ s|<.*?>||g;
    $line =~ s/</&lt;/g;
    $line =~ s/>/&gt;/g;
    $line =~ s/"/&quot;/g;
    $line;
}

##
## Remove strange characters (non-ASCII, non-EUC)
##
sub remove_strange_chars {
    my ($line) = @_;
    my $r = '';
    while ($line =~ /([\x20-\x7e]|\x8f?[\xa1-\xfe][\xa1-\xfe])+/g) {
	$r .= $&;
    }
    $r;
}

##
## Insert spaces between characters
##
sub insert_spaces {
    my ($line) = @_;
    my @line = split(//, $line);
    @_ = ();
    while (@line) {
	$_ = shift(@line);
	if ($_ eq "\x8f") {
	    $_ .= shift(@line);
	    $_ .= shift(@line);
	} elsif ("\x80" le $_ && $_ le "\xff") {
	    $_ .= shift(@line);
	}
	push(@_, $_);
    }
    return join(" ", @_);
}

##
## GMT string
##
sub GMT_string {
    my ($time) = @_;
    @_ = split(/\s+/, scalor(gmtime($time)));
    sprintf "%s, %02d %s %s %s GMT", $_[0], $_[2], $_[1], $_[4], $_[3];
}

################################################################
##
## Search engines
##
################################################################
sub init_engine {
    my ($eng, $max_hits) = @_;
    my $domain = &url_encode($param{'domain'});
    my $q_euc = $param{'q_euc'};
    my $q_sjis = $param{'q_sjis'};
    my ($server, $ip, $port, $r, $extract);
    $port = 80;
    $max_hits = 100 if $max_hits > 100;
    if ($eng eq 'altavista') {
	$server = 'www.altavista.com';
	$ip = '66.94.229.254';
        $r = '/web/results?q=' . $q_sjis;
	if ($domain) {
	    $r .= '&pg=ps&rc=url&lh=' . $domain;
	} else {
	    $r .= '&pg=q';
	}
	$r .= '&kl=ja&enc=sjis';
	$r .= '&nbq=' . $max_hits;
	$extract = \&altavista;
    } elsif ($eng eq 'excite') {
	$server = 'www.excite.co.jp';
	$ip = '210.150.160.234';
	$max_hits = 20 if $max_hits > 20;
	$r = '/search.gw?lk=excite_jp&lang=jp&c=web&';
	$r .= 'search=' . $q_sjis;
	if ($domain) {
	    $r .= '+%2Bsite:' . $domain;
	}
	$r .= '&NumResults=' . $max_hits;
	$extract = \&excite;
    } elsif ($eng eq 'fast') {
	$server = 'www.alltheweb.com';
	$ip = '66.94.229.254';
	$r = '/search?cat=web&cs=cp932&_sb_lang=pref&';
	$r .= 'q=' . $q_sjis;
	$r .= '&hits=' . $max_hits;
	if ($domain) {
	    $r .= '&advanced=1&dincl=' . $domain;
	}
	$extract = \&fast;
    } elsif ($eng eq 'fresheye') {
	$server = 'search.fresheye.com';
	$ip = '210.174.177.241';
	$r = '/?kw=' . $q_euc;
	$extract = \&fresheye;
    } elsif ($eng eq 'goo') {
	$server = 'search.goo.ne.jp';
	$ip = '210.150.25.240';
	$max_hits = 20 if $max_hits == 30;
	$r = '/web.jsp?MT=' . $q_euc;
	$r .= '&SM=MC&DC=' . $max_hits;
	$r .= '&SDB=PG';
	if ($domain) {
	    $r .= '&RD=DM&DS=IN&Domain='. $domain;
	}
	$extract = \&goo;
    } elsif ($eng eq 'infoseek') {
	$server = 'search.www.infoseek.co.jp';
	$ip = '202.232.118.11';
	$r = '/Seek?qt=' . $q_euc;
	if ($domain) {
	    $r .= '+%2Bsite:' . $domain;
	}
	$r .= '&nh=' . $max_hits;
	$extract = \&infoseek;
    } elsif ($eng eq 'verno') {
	$server = 'verno.ueda.info.waseda.ac.jp';
	$ip = '133.9.215.87';
	$r = '/verno-099.cgi?index-and=' . $q_euc;
	$r .= '&kosuu=' . $max_hits;
	$r .= '&dig=-1';
	$extract = \&verno;
    } elsif ($eng eq 'yahoo') {
	$server = 'search.yahoo.co.jp';
	$ip = '203.216.227.246';
	$max_hits = 100 if $max_hits == 50;
	$r = '/bin/query?p=' . $q_euc;
	if ($domain) {
	    $r .= '+%2Bsite:' . $domain;
	}
	$r .= '&n=' . $max_hits;
	$extract = \&yahoo;
    }
    my $proxy_server = $const{'proxy_server'};
    my $proxy_port = $const{'proxy_port'};
    my $no_proxy = $const{'no_proxy'};
    if ($proxy_server && $proxy_port && $server !~ /$no_proxy/) {
	$r = "http://$server" . $r;
	$server = $proxy_server;
	undef $ip;
	$port = $proxy_port;
    }
    $eng{$eng}{'server'} = $server;
    $eng{$eng}{'ip'} = $ip;
    $eng{$eng}{'port'} = $port;
    $eng{$eng}{'request'} = $r;
    $eng{$eng}{'extract'} = $extract;
}

sub altavista {
    my ($eng, $content) = @_;
    my ($re, $href, $title, $score, $comment);
    my $max_hits = $param{'max_hits'};
    &jcode::convert(\$content, 'euc', 'sjis');
    unless ($content =~ m|^.*?<br class=\'lb\'>|i) {
	return;
    }
    $content = $';
    $content =~ s|<b class=m>.*$||i;
    $re = '<a class=\'res\' href=\'(http.*?)\'>';
    $re .= '(.*?)</a><br>(.*?)<br>';
    my $n = 0;
    while ($content =~ /$re/igo) {
	$href = $1;
	$title = $2;
	$comment = $3;
	$score = int(100 - 30*$n/$max_hits);
	&add_url($eng, $href, $title, $score, $comment);
	$n++;
    }
}

sub excite {
    my ($eng, $content) = @_;
    my ($re, $href, $title, $score, $comment);
    my $max_hits = $param{'max_hits'};
    &jcode::convert(\$content, 'euc', 'sjis');
    unless ($content =~ m|^.*?<div class="title"><FONT SIZE=-1><B>ָ</B>|i) {
	return;
    }
    $content =~ s|<br><br>.*$||i;
    $re = '<div class="hit">\s*<a href="(http.*?)".*?>(.*?)</a>\s*';
    $re .= '.*?<table.*?><tr><td.*?><font.*?>(.*?)</font>';
    my $n = 0;
    while ($content =~ /$re/igo) {
	$href = $1;
	$title = $2;
	$comment = $3;
	$score = int(100 - 30*$n/$max_hits);
	&add_url($eng, $href, $title, $score, $comment);
	$n++;
    }
}

sub fast {
    my ($eng, $content) = @_;
    my ($re, $href, $title, $score, $comment);
    my $max_hits = $param{'max_hits'};
    &jcode::convert(\$content, 'euc', 'sjis');
    unless ($content =~ m|^.*<!--RS-->|i) {
	return;
    }
    $content = $';
    $content =~ s|<!--RE-->.*$||i;
    $re = '<a class="res" href="(.*?)"\s+>(.*?)</a>.*?';
    $re .= '<span class="resTeaser">(.*?)</span.';
    my $n = 0;
    while ($content =~ /$re/igo) {
	$href = $1;
	$title = $2;
	$comment = $3;
	$score = int(100 - 30*$n/$max_hits);
	&add_url($eng, $href, $title, $score, $comment);
	$n++;
    }
}

sub fresheye {
    my ($eng, $content) = @_;
    my ($re, $href, $title, $score, $comment);
    my $max_hits = $param{'max_hits'};
    &jcode::convert(\$content, 'euc', undef);
    unless ($content =~ m|<ol start=1>|i) {
	return;
    }
    $content = $';
    $content =~ s|</ol>||i;
    $re = '<li>\s*<p class="title"><a href=.*?>(.*?)</a>\s*</p>\s*';
    $re .= '<p class="brief">(.*?)</p>\s*<p class="url">(http://\S+)';
    my $n = 0;
    while ($content =~ /$re/igo) {
	$href = $3;
	$title = $1;
	$comment = $2;
	$score = int(100 - 30*$1/$max_hits);
	&add_url($eng, $href, $title, $score, $comment);
	$n++;
    }
}

sub goo {
    my ($eng, $content) = @_;
    my ($re, $href, $title, $score, $comment);
    my $max_hits = $param{'max_hits'};
    &jcode::convert(\$content, 'euc', undef);
    unless ($content =~ m|^.*?<!-- google search result -->|i) {
	return;
    }
    $content = $';
    $content =~ s|<br clear.*$||i;
    $re = '(\d+)&nbsp;\s*<a href="(.*?)".*?>(.*?)</a><br>\s*';
    $re .= '<div class="r">\s*<font.*?>(.*?)<div class="g">';
    while ($content =~ /$re/igo) {
	$href = $2;
	$title = $3;
	$comment = $4;
	$href =~ s/%3F/?/ig;
	$href =~ s/%3D/=/ig;
	$href =~ s/%26/&/ig;
	$score = int(100 - 30*$1/$max_hits);
	&add_url($eng, $href, $title, $score, $comment);
    }
}

sub infoseek {
    my ($eng, $content) = @_;
    my ($re, $href, $title, $score, $comment);
    my $max_hits = $param{'max_hits'};
    &jcode::convert(\$content, 'euc', undef);
    unless ($content =~ m|^.*?<!-- START RESULT LIST -->|i) {
	return;
    }
    $content = $';
    $content =~ s|<!-- END RESULT LIST -->.*$||i;
    $re = '<!-- \d+ -->\s*<li>\s*<div.*?>\s*';
    $re .= '<a href="(.*?)".*?>(.*?)</a>\s*</div>\s*';
    $re .= '<div.*?><small>(.*?)</small></div>\s*.*?</li>';
    my $n = 0;
    while ($content =~ /$re/igo) {
	$href = $1;
	$title = $2;
	$score = int(100 - 30*$n/$max_hits);
	$comment = $3;
	&add_url($eng, $href, $title, $score, $comment);
	$n++;
    }
}

sub yahoo {
    my ($eng, $content) = @_;
    my ($re, $href, $title, $score, $comment);
    my $max_hits = $param{'max_hits'};
    &jcode::convert(\$content, 'euc', 'sjis');
    unless ($content =~ m|^.*?<div class="web">|i) {
	return;
    }
    $content = '<div class="web">' . $';
    # $content =~ s|</ol>.*$||i;
    $re = '<div class="web">.*?<h3><a.*?http%3A(.*?)">(.*?)</a>.*?</div>';
    $re .= '\s*<div class="bd"><div class="abs">(.*?)</div>';
    my $n = 0;
    while ($content =~ /$re/igo) {
	$href = "http:$1";
	$title = $2;
	$comment = $3;
	$score = int(100 - 30*$n/$max_hits);
	&add_url($eng, $href, $title, $score, $comment);
	$n++;
	last if $n > $max_hits;
    }
}
